/*
 *	sms object  (Apple Sudden Motion Sensor) 
 * 
 *		Copyright(C) 2006-2007 Shigeyuki Hirai
 */


#include "ext.h"
#include <IOKit/IOKitLib.h>

struct data {
	char x;
	char y;
	char z;
	char pad[57];
};

typedef struct sms
{
	t_object m_ob;
	void *m_xout;
	void *m_yout;
	void *m_zout;
	void *m_clock;
	double m_interval;
	long m_run;
} t_sms;

void AccessSMS(t_sms *x);
void sms_float(t_sms *x, double f);
void sms_int(t_sms *x, long n);
void sms_tick(t_sms *x);
void sms_bang(t_sms *x);
void sms_start(t_sms *x);
void sms_stop(t_sms *x);
void sms_rate(t_sms *x, long rate);
void sms_free(t_sms *x);
void *sms_new(t_symbol *s, short argc, t_atom *argv);

void *sms_class;

void AccessSMS(t_sms *x)
{
	kern_return_t result;
	mach_port_t masterPort;
	io_iterator_t iterator;
	io_object_t aDevice;
	io_connect_t  dataPort;
	int kernFunc;
	int dataSize;
	
	IOItemCount structureInputSize;
	IOByteCount structureOutputSize;

	struct data inputStructure;
	struct data outputStructure;

	result = IOMasterPort(MACH_PORT_NULL, &masterPort);
	if (result != KERN_SUCCESS) {
		error("sms: IOMasterPort(MACH_PORT_NULL, &masterPort)");
		return;
	}

	//PowerBookG4, iBookG4
	result = IOServiceGetMatchingServices(masterPort, IOServiceMatching("IOI2CMotionSensor"), &iterator);
	if (result == KERN_SUCCESS && iterator != 0) {
		dataSize = 60;
		kernFunc = 21;
		goto FOUND_SENSOR;
	}
	//
	result = IOServiceGetMatchingServices(masterPort, IOServiceMatching("PMUMotionSensor"), &iterator);
	if (result == KERN_SUCCESS && iterator != 0) {
		dataSize = 60;
		kernFunc = 21;
		goto FOUND_SENSOR;
	}
	//
	result = IOServiceGetMatchingServices(masterPort, IOServiceMatching("SMCMotionSensor"), &iterator);
	if (result == KERN_SUCCESS && iterator != 0) {
		dataSize = 40;
		kernFunc = 5;
		goto FOUND_SENSOR;
	}
	
	error("can't find motionsensor\n");
	return;


FOUND_SENSOR:
	aDevice = IOIteratorNext(iterator);
	IOObjectRelease(iterator);
	
	if (aDevice == 0) {
		error("sms: No motion sensor available.");
		return;
	}

	result = IOServiceOpen(aDevice, mach_task_self(), 0, &dataPort);
	IOObjectRelease(aDevice);

	if (result != KERN_SUCCESS) {
		error("sms: Could not open motion sensor device.");
		return;
	}


	structureInputSize = dataSize;	//sizeof(struct data);
	structureOutputSize = dataSize;	//sizeof(struct data);

	memset(&inputStructure, 0, sizeof(inputStructure));
	memset(&outputStructure, 0, sizeof(outputStructure));

	result = IOConnectMethodStructureIStructureO(
				dataPort,
				kernFunc,	/* index to kernel function */
				structureInputSize,
				&structureOutputSize,
				&inputStructure,
				&outputStructure);

	if (result == KERN_SUCCESS) {
//		post("%d %d %d", outputStructure.x, outputStructure.y, outputStructure.z);
		outlet_float(x->m_zout, (float)outputStructure.z);
		outlet_float(x->m_yout, (float)outputStructure.y);
		outlet_float(x->m_xout, (float)outputStructure.x);
	} else {
		error("sms: no coords");
	}

	IOServiceClose(dataPort);
	return;
}

void main()
{
	setup((t_messlist **)&sms_class, 
			(method)sms_new, 
			(method)sms_free, 
			(short)sizeof(t_sms), 
			0L, 
			A_GIMME, 0);
	addint((method)sms_int);
	addfloat((method)sms_float);
	addbang((method)sms_bang);
	addmess((method)sms_start, "start", 0);
	addmess((method)sms_stop, "stop", 0);
	addmess((method)sms_rate, "rate", A_LONG, 0);
}

void sms_float(t_sms *x, double f)
{
	sms_int(x, (long)f);
}

void sms_tick(t_sms *x)
{
	AccessSMS(x);
	if (x->m_run)
		clock_fdelay(x->m_clock, x->m_interval);
}

void sms_bang(t_sms *x)
{
	AccessSMS(x);
}

void sms_start(t_sms *x)
{
	sms_int(x, 1);
}

void sms_stop(t_sms *x)
{
	sms_int(x, 0);
}

void sms_int(t_sms *x, long n)
{
	if (n != 0) {
		if (x->m_run == 0) {
			x->m_run = 1;
			clock_delay(x->m_clock, 0);
		}
	} else {
		x->m_run = 0;
	}
}

void sms_rate(t_sms *x, long rate)
{
	if (rate > 0) {
		x->m_interval = 1000./rate;
	}
}

void sms_free(t_sms *x)
{
	freeobject(x->m_clock);
}

void *sms_new(t_symbol *s, short argc, t_atom *argv)
{
	t_sms *x = (t_sms *)newobject(sms_class);
	x->m_zout = floatout(x);
	x->m_yout = floatout(x);
	x->m_xout = floatout(x);
	x->m_clock = clock_new(x, (method)sms_tick);
	x->m_run = 0;
	x->m_interval = 100.;
	return x;
}
